////////////////////////////////////////////////////////////////////////////////
// 
// LPinter.cc
//
//    produced: 2004/04/28 cw
//
// This program is free software; Please read COPYING
//  (GNU General Public Licence) for details
//  this file was adapted from the file LPinterface by Joerg Rambau for TOPCOM
//
////////////////////////////////////////////////////////////////////////////////

#include "LPinter.hh"

// constructors:

LPinter::LPinter(const Matrix& m, const Vector &v)
{
  if (CommandlineOptions::debug()) {
    std::cerr << "building LP matrix ..." << std::endl;
  }

  _solver                     = dd_DualSimplex;
  
  // note that the coefficient matrix m is transposed for efficiency reasons;
  _m                          = m.coldim(); // number of rows
  _n                          = m.rowdim(); // number of cols
  _matrixptr                  = dd_CreateMatrix(_m, _n);
  _matrixptr->objective       = dd_LPmax;
  _matrixptr->numbtype        = dd_Rational;

  for (size_type j = 0; j < m.rowdim(); ++j) {

    // build cdd's matrix row by row;
    // recall that the coefficient matrix m is transposed for efficiency reasons;
    // build objective vector too
    dd_bis_set_R(_matrixptr->rowvec[j], v[j]);

    for (size_type i = 0; i < m.coldim(); ++i) {

      // because the first column contains the data for the right hand side 
      // (which is ZERO) we need to shift the column index by one;
      // again, we have to use the coefficient matrix m in a transposed way:
      dd_bis_set_R(_matrixptr->matrix[i][j], m(j, i));
    }
  }
  if (CommandlineOptions::debug()) {
    dd_WriteMatrix(stderr, _matrixptr);
  }

  // cdd allows to automatically generate an auxilliary LP for finding an interior point:
  _mainlpptr = dd_Matrix2LP(_matrixptr, &_err);

  if (CommandlineOptions::debug()) {
    dd_WriteLP(stderr, _mainlpptr);
  }


  _lpptr = dd_MakeLPforInteriorFinding(_mainlpptr);
  if (!_lpptr) {
    std::cerr << "error in generating LP." << std::endl;
    dd_WriteErrorMessages(stderr, _err);
    exit(1);
  }
  if (CommandlineOptions::debug()) {
    std::cerr << "... done." << std::endl;
  }
}

// functions:
Vector LPinter::solve() {

  // solve problem
  const bool result2 = dd_LPSolve(_mainlpptr, _solver, &_err);

  if (_err != dd_NoError) {
    std::cerr << "error in solving LP." << std::endl;
    dd_WriteErrorMessages(stderr, _err);
    exit(1);
  }


  if (_mainlpptr->re>0) {
    if (CommandlineOptions::debug()) {
      std::cerr << "No solution, certificate: " << _mainlpptr->re << std::endl;
    }
    Vector result = Vector();
    return result;
  }

  // Write problem

  if (CommandlineOptions::debug()) {
    dd_WriteLPResult(stderr, _mainlpptr, _err);

    if (_err != dd_NoError) {
      std::cerr << "error in writing LP result." << std::endl;
      dd_WriteErrorMessages(stderr, _err);
      exit(1);
  }

  }

  // Write solution
  dd_LPSolutionPtr auxsolptr = dd_CopyLPSolution(_mainlpptr);
  Vector result = Vector(auxsolptr->d);

  for (dd_rowrange j = 0; j < (auxsolptr->d); j++) {
    result[j] = Field(auxsolptr->sol[j]);
  }

  dd_FreeLPSolution(auxsolptr);

  return result;
}

bool LPinter::has_interior_point() {
  
  // Check feasibility with cdd LP library:
  const bool result = dd_LPSolve(_lpptr, _solver, &_err);
  if (_err != dd_NoError) {
    std::cerr << "error in solving LP." << std::endl;
    dd_WriteErrorMessages(stderr, _err);
    exit(1);
  }

  // Write an interior point:
  _solptr = dd_CopyLPSolution(_lpptr);
  if (dd_Positive(_solptr->optvalue)){
    if (CommandlineOptions::debug()) {
      std::cout << "(";
      Field maxheight(ZERO);
      for (dd_rowrange j = 0; j < (_solptr->d) - 2; j++) {
	const Field x_j = Field(_solptr->sol[j+1]);
	if (maxheight < x_j) {
	  maxheight = x_j;
	}
      }
      for (dd_rowrange j = 0; j < (_solptr->d) - 3; j++) {
	const Field x_j = Field(_solptr->sol[j+1]);
	std::cout << double(x_j);
	std::cout << ",";
      }
      const long j((_solptr->d) - 3);
      const Field x_j = Field(_solptr->sol[j+1]);
      std::cout << double(x_j);
      std::cout << ")" << std::endl;
      if (CommandlineOptions::debug()) {
	std::cerr << "Optimal slack: ";
	dd_WriteNumber(stderr, _solptr->optvalue);
	std::cerr << std::endl;
      }
    }
    if (CommandlineOptions::debug()) {
      std::cerr << "... done." << std::endl;
    }
    return true;
  }
  if (dd_Negative(_solptr->optvalue)) {
    if (CommandlineOptions::debug()) {
      std::cerr << "The feasible region is empty." << std::endl;
    }
    if (CommandlineOptions::debug()) {
      std::cerr << "... done." << std::endl;
    }
    return false;
  }
  if (dd_EqualToZero(_solptr->optvalue)) {
    if (CommandlineOptions::debug()) {
      std::cerr << "... done." << std::endl;
    }
    if (CommandlineOptions::debug()) {
      std::cerr << "The feasible region is nonempty but has no interior point." << std::endl;
    }
    return false;
  }
  return false;
}

// eof LPinter.cc
