////////////////////////////////////////////////////////////////////////////////
// 
// VPolytope.cc
//
//    produced: 21/05/04 cw
// last change: 21/05/04 cw
// 
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <fstream>
#include <assert.h>
#include <ctype.h>
#include <string.h>

#include "VPolytope.hh"


// the following preprocessor magic is necessary because the word `Rational'
// is used in an enum type in the cddlib:

// we can use three variants of arithmetics in cddlib:

#if defined GMPRATIONAL // GMP Rational
  
inline void dd_bis_set_R(mytype& a, const Rational& b) {
  //  mpq_init(a);
  mpq_set(a, b.get_rep());
}

#elif defined GMPFLOAT // GMP float
  
inline void dd_bis_set_R(mytype& a, const Rational& b) {
  //  mpq_init(a);
  mpq_set(a, double(b));
}

#else // built-in C double 
  
inline void dd_bis_set_R(mytype& a, const Rational& b) {
  a[0] = double(b);
}

#endif


void VPolytope::writeAdjacency(std::ostream& ost) {
  ost << _adj << std::endl;
}

std::istream& VPolytope::readAdjacency(std::istream& ist) {
  return ist >> _adj;
}

void VPolytope::computeAdjacency() {
  dd_MatrixPtr ddmatrixptr = dd_CreateMatrix(coldim(),rowdim()+1);
  ddmatrixptr->representation = dd_Generator;
  dd_ErrorType dderr;

  for (size_type i = 0; i < coldim(); ++i) {
    // build cdd's matrix row by row
    // recall that the coefficient matrix m is transposed for efficiency reasons

    // because the first column contains a one
    dd_bis_set_R(ddmatrixptr->matrix[i][0], 1);

    for (size_type j = 0; j < rowdim(); ++j) {
      dd_bis_set_R(ddmatrixptr->matrix[i][j+1], (*this)[i][j]);
    }
  }

  if (CommandlineOptions::debug()) {
    dd_WriteMatrix(stderr, ddmatrixptr);
  }

  dd_SetFamilyPtr ddsetfamilyptr = dd_Matrix2Adjacency(ddmatrixptr, &dderr);

  if (dderr != dd_NoError) {
    std::cerr << "error in computing adjacency list." << std::endl;
    dd_WriteErrorMessages(stderr, dderr);
    exit(1);
  }

  _adj = Adjacency();

  for (dd_rowrange j = 0; j < (ddsetfamilyptr->famsize); j++) {

    IntegerVector list = IntegerVector();
    for (parameter_type i = 1; i <= (ddsetfamilyptr->setsize); i++) {
      if (set_member(i,ddsetfamilyptr->set[j])) {
	list.append(i-1);
      }
    }
    _adj.append(list);
  }


  dd_FreeSetFamily(ddsetfamilyptr);
  dd_FreeMatrix(ddmatrixptr);
}

void VPolytope::removeRedundancy() {
  dd_MatrixPtr ddmatrixptr = dd_CreateMatrix(coldim(),rowdim()+1);
  ddmatrixptr->representation = dd_Generator;
  dd_ErrorType dderr;

  for (size_type j = 0; j < rowdim(); ++j) {

    // build cdd's matrix row by row;

    for (size_type i = 0; i < coldim(); ++i) {

      dd_bis_set_R(ddmatrixptr->matrix[i][j+1], (*this)[i][j]);
    }
  }

  for (size_type i = 0; i < coldim(); ++i) {
    // because the first column contains a one
    dd_bis_set_R(ddmatrixptr->matrix[i][0], 1);
  }

  if (CommandlineOptions::debug()) {
    dd_WriteMatrix(stderr, ddmatrixptr);
  }
  dd_rowset ddredrows=dd_RedundantRows(ddmatrixptr, &dderr);
  if (dderr != dd_NoError) {
    std::cerr << "error when removing redundant vertices." << std::endl;
    dd_WriteErrorMessages(stderr, dderr);
    exit(1);
  }

  dd_rowrange msub = ddmatrixptr->rowsize;
  dd_rowrange m = msub;
  for (dd_rowrange i=1; i<=m; i++) {
    if (set_member(i,ddredrows))
      msub-=1;
  }
  
  resize(msub);

  bool same = true;
  size_type diff = 0;

  for (dd_rowrange i = 0; i < m; i++) {

    if (set_member(i+1,ddredrows)) {
      same = false;
      diff++;
      continue;
    }
    if (same) continue;

    // build poly's matrix row by row, starting from the first changed row
    // recall that the coefficient matrix m is transposed for efficiency reasons;
    for (size_type j = 0; j < rowdim(); ++j) {
      (*this)[i-diff][j] = Field(ddmatrixptr->matrix[i][j+1]);
    }
  }
  dd_FreeMatrix(ddmatrixptr);
}

Vector VPolytope::fej(const parameter_type& vtx_no,const parameter_type& ngbr_no) {
  assert(vtx_no>=0);
  assert(ngbr_no>=0);
  if(vtx_no>=no())
    return Vector();

  if (ngbr_no >= ngbrCount(vtx_no))
    return Vector();
  Vector result = Vector((*this)[vtx_no]);
  result.scale(-1);
  result.add((*this)[faj(vtx_no,ngbr_no)]);
  return result;
}

Vector VPolytope::fej0(const parameter_type& vtx_no,const parameter_type& ngbr_no) {
  Vector result = Vector(1);
  result.append(fej(vtx_no, ngbr_no));
  return result;
}

const size_type VPolytope::ngbrCount(const parameter_type& vtx_no) {
 
  return _adj[vtx_no].maxindex();
}

// eof VPolytope.cc
