////////////////////////////////////////////////////////////////////////////////
// 
// VPolytopeList.cc
//
//    produced: 22/11/05 cw
// last change: 22/11/05 cw
// 
// This program is free software; Please read COPYING
//  (GNU General Public Licence) for details
//  this file was written by Christophe Weibel for MINKSUM
//
////////////////////////////////////////////////////////////////////////////////
#include <iostream>
#include <fstream>
#include <ctype.h>
#include <string.h>

#include "VPolytopeList.hh"

// NB : index of polytopes are from 0 to numPol()-1
// NB : index of neighbours in adjacency are from 1 to ngbrCount()

const double EPSILON = 0.000000001;

// returns a maximizer for this vertex
Vector VPolytopeList::maximizer(const IntegerVector current) {
  // std::cout << "max " << current << std::endl;
  Vector objV = Vector((*this)[0].rank()+1);
  Matrix pbV = Matrix((*this)[0].rank()+1,0);

  for (int j=0; j<numPol(); ++j)
    for (int i=0; i < (*this)[j].ngbrCount(current[j]); ++i) {

      Vector toAdd = (*this)[j].fej0(current[j],i);
      objV.add(toAdd);
      toAdd.scale(-1);
      toAdd[0]=-1;
      pbV.append(toAdd);
    }
  if (CommandlineOptions::positive()) {
    // only compute maximizers in the positive quadrant
    for (int l=1; l<=(*this)[0].rank();++l) {
      Vector toAdd = Vector((*this)[0].rank()+1,0);
      toAdd[l] = 1;
      pbV.append(toAdd);
    }
  }

  LPinter LPV(pbV,objV);

  Vector maximizer = LPV.solve();

  return maximizer;
}

// returns the vertex decomposition
// of the adjacent vertex along this edge
IntegerVector VPolytopeList::neighbour(const IntegerVector cur_vtx, const int poly, const int ngbr_no) {
  IntegerVector result = IntegerVector((*this).numPol(),0);

  // first polytope has index 0
  for (int j=0; j < (*this).numPol(); ++j) {
    result[j] = cur_vtx[j];
    // first neighbour has index 0
    for (int i=0; i<(*this)[j].ngbrCount(cur_vtx[j]); ++i) {
      if ((*this)[j].fej(cur_vtx[j],i).
	  is_colinear((*this)[poly].fej(cur_vtx[poly],ngbr_no))) {
	result[j] = (*this)[j].faj(cur_vtx[j],i);
      }
    }
  }

  return result;
}

//returns whether there is an adjacent vertex along this edge
bool VPolytopeList::adj(const  IntegerVector cur_vtx, const int poly, const int ngbr_no) {
  Matrix pb = Matrix((*this)[0].rank(),0);

  Vector ob,lim;

  ob = ((*this)[poly]).fej0(cur_vtx[poly],ngbr_no);
  lim = Vector(ob);
  lim[0]=1;
  pb.append(lim);

  ob.scale(-1);

  // first polytope has index 0
  for (int j=0; j < (*this).numPol(); ++j) {
    // first neighbour has index 1
    for (int i=0; i<(*this)[j].ngbrCount(cur_vtx[j]); ++i) {
      if (!(*this)[j].fej(cur_vtx[j],i).
	  is_colinear((*this)[poly].fej(cur_vtx[poly],ngbr_no))) {
	Vector add = (*this)[j].fej0(cur_vtx[j],i);
	pb.append(add);
      } else if (j<poly || (j==poly&&i<ngbr_no)) {
	return false;
      }
    }
  }

  LPinter LP(pb,ob);
  Vector result = LP.solve();

  return inner_product(result, ob) > 0;
}

// Finds out the parent of this vertex, with this maximizer for efficiency.
// Returns an empty vector if it is the root
IntegerVector VPolytopeList::parent(IntegerVector child, Vector childMax) {
  // std::cout << "parent " << child << childMax << std::endl;
  Vector diff = Vector(childMax);
  diff.scale(-1);
  diff.add(obj);
  Matrix pbV = Matrix(childMax.maxindex(),0);

  for (int l=0; l<(*this).numPol(); ++l)
    for (int k=0; k<(*this)[l].ngbrCount(child[l]); ++k)
      if (inner_product((*this)[l].fej0(child[l],k),diff)>0) {
	Vector toAdd = Vector((*this)[l].fej0(child[l],k));
	toAdd.scale(-1);
	pbV.append(toAdd);
      }

  RSinter RSV = RSinter(pbV,childMax,diff);
  int limit = RSV.shoot();

  if (limit < 1) {
    return IntegerVector();
  }

  int firstPol = 0;
  int firstNgbr = 0;
  bool cont = true;

  // find out which is the parent for child vertex
  for (int l=0; cont && l<(*this).numPol(); ++l)
    for (int k=0; cont && k<(*this)[l].ngbrCount(child[l]); ++k)
      if (inner_product((*this)[l].fej0(child[l],k),diff)>0) {
	if (--limit == 0) {
	  cont = false;
	  firstPol = l;
	  firstNgbr = k;
	}
      }
  return (*this).neighbour(child,firstPol,firstNgbr);
}

// checks if more vertices remain
bool VPolytopeList::hasMoreVertices() {
  return vertex.maxindex() != 0;
}

// does one step of exploration of the vertices of the minkowski sum
// computes facets of the normal cone, but is slower
SumVertex VPolytopeList::incExploreStep() {
  if (CommandlineOptions::verbose()) {
    std::cerr << "IncExploreStep " << vertex << std::endl;
  }

  IntegerVector next;
  Vector nextMax;

  SumVertex result;
  VPolytope facets = VPolytope();

  if (CommandlineOptions::debug()) {
    std::cerr << ".";
  }

  bool found = false;

  for (int j=0; j<(*this).numPol(); ++j)
    for (int i=0;
	 i<(*this)[j].ngbrCount(vertex[j]);
	 ++i) { // loop on leaving edges
	
	
      if ((*this).adj(vertex,j,i)) { // leaving edge exists
	Vector edge = (*this)[j].fej0(vertex[j],i);
	edge.scale(-1);
	facets.append(edge);

	// attempts to find out the next child vertex, if not yet found
	if (!found && inner_product(edge,obj)>0) { // examine only away from objective
	  IntegerVector test = (*this).neighbour(vertex,j,i);

	  Vector testMax = (*this).maximizer(test);
	  if (testMax.maxindex()==0) { // No satisfactory maximizer found
	    // actually, this sould only happen if we choose
	    // to forbid maximizers with negative values
	    if (!CommandlineOptions::positive()) {
	      std::cerr << "ERROR : THIS IS A BUG" << std::endl;
	    }
	    continue; // don't examine point at all.
	  }

	  if (CommandlineOptions::verbose()) {
	    std::cerr << "testing(test,testmax)" << test << "," << testMax << std::endl;
	  }

	  if (parent(test,testMax)==vertex) { // check parent
	    found = true;
	    next = test;
	    nextMax = testMax;
	  } // check parent
	} // examine only away from objective
      } // leaving edge exists
    } // loop on leaving edges


  Vector coord = (*this).output(vertex);
  result = SumVertex(coord,vertex,vertexMax,facets);

  while (!found) { // Go up the tree to look for next vertex in other branches
    IntegerVector justVisited = vertex;

    if(vertexMax == obj) { // no more vertices
      vertex = IntegerVector();
      return result;
    }
    vertex = parent(vertex,vertexMax);
    vertexMax = maximizer(vertex);
    Vector newcoord = (*this).output(vertex);
    coord.scale(-1);
    coord.add(newcoord); // coord now contains the last edge

    bool visitedBefore = true;
    for (int j=0; j<(*this).numPol(); ++j)
      for (int i=0;
	   i<(*this)[j].ngbrCount(vertex[j]);
	   ++i) { // loop on leaving edges
	Vector edge = (*this)[j].fej(vertex[j],i);
	if (visitedBefore) { // don't test vertices already visited
	  if (edge.is_colinear(coord)) { // test from now on
	    visitedBefore = false;
	  }
	} else { // don't test/test
	  edge = (*this)[j].fej0(vertex[j],i);
	  // attempts to find out the next child vertex
	  if (inner_product(edge,obj)<0 &&
	      (*this).adj(vertex,j,i)) { // examine only away from objective
	    IntegerVector test = (*this).neighbour(vertex,j,i);
	    Vector testMax = (*this).maximizer(test);

	    if (testMax.maxindex()==0) { // No satisfactory maximizer found
	      // actually, this should only happen if we choose
	      // to forbid maximizers with negative values
	      if (!CommandlineOptions::positive()) {
		std::cerr << "ERROR : THIS IS A BUG" << std::endl;
	      }
	      continue; // don't examine point at all.
	    }

	    if (CommandlineOptions::verbose()) {
	      std::cerr << "testing(test,testmax)" << test << "," << testMax << std::endl;
	    }

	    if (parent(test,testMax)==vertex) { // check parent
	      vertex = test;
	      vertexMax = testMax;
	      return result;
	    } // check parent
	  } // examine only away from objective
	} // test
      } // loop on leaving edges

    coord = newcoord; // save the current coordinates
  } // while

  vertex = next;
  vertexMax = nextMax;
  return result;
}

// does one step of exploration of the vertices of the minkowski sum fast
// does not computes facets of the normal cone
SumVertex VPolytopeList::incExploreStepFast() {
  if (CommandlineOptions::verbose()) {
    std::cerr << "IncExploreStep " << vertex << std::endl;
  }

  IntegerVector next;
  Vector nextMax;

  SumVertex result;
  VPolytope facets = VPolytope();

  if (CommandlineOptions::debug()) {
    std::cerr << ".";
  }

  bool found = false;

  for (int j=0; (!found) && j<(*this).numPol(); ++j)
    for (int i=0;
	 (!found) && i<(*this)[j].ngbrCount(vertex[j]);
	 ++i) { // loop on leaving edges
      Vector edge = (*this)[j].fej0(vertex[j],i);

      // attempts to find out the next child vertex, if not yet found
      if (inner_product(edge,obj)<0 &&
	  (*this).adj(vertex,j,i)) { // examine only away from objective
	IntegerVector test = (*this).neighbour(vertex,j,i);

	Vector testMax = (*this).maximizer(test);
	if (testMax.maxindex()==0) { // No satisfactory maximizer found
	  // actually, this sould only happen if we choose
	  // to forbid maximizers with negative values
	  if (!CommandlineOptions::positive()) {
	    std::cerr << "ERROR : THIS IS A BUG" << std::endl;
	  }
	  continue; // don't examine point at all.
	}

	if (CommandlineOptions::verbose()) {
	  std::cerr << "testing(test,testmax)" << test << "," << testMax << std::endl;
	}

	if (parent(test,testMax)==vertex) { // check parent
	  found = true;
	  next = test;
	  nextMax = testMax;
	} // check parent
      } // examine only away from objective
    } // loop on leaving edges


  Vector coord = (*this).output(vertex);
  result = SumVertex(coord,vertex,vertexMax,facets);

  while (!found) { // Go up the tree to look for next vertex in other branches
    IntegerVector justVisited = vertex;

    if(vertexMax == obj) { // no more vertices
      vertex = IntegerVector();
      return result;
    }
    vertex = parent(vertex,vertexMax);
    vertexMax = maximizer(vertex);
    Vector newcoord = (*this).output(vertex);
    coord.scale(-1);
    coord.add(newcoord); // coord now contains the last edge

    bool visitedBefore = true;
    for (int j=0; j<(*this).numPol(); ++j)
      for (int i=0;
	   i<(*this)[j].ngbrCount(vertex[j]);
	   ++i) { // loop on leaving edges
	Vector edge = (*this)[j].fej(vertex[j],i);
	if (visitedBefore) { // don't test vertices already visited
	  if (edge.is_colinear(coord)) { // test from now on
	    visitedBefore = false;
	  }
	} else { // don't test/test
	  edge = (*this)[j].fej0(vertex[j],i);
	  // attempts to find out the next child vertex
	  if (inner_product(edge,obj)<0 &&
	      (*this).adj(vertex,j,i)) { // examine only away from objective
	    IntegerVector test = (*this).neighbour(vertex,j,i);
	    Vector testMax = (*this).maximizer(test);
	    if (testMax.maxindex()==0) { // No satisfactory maximizer found
	      // actually, this should only happen if we choose
	      // to forbid maximizers with negative values
	      if (!CommandlineOptions::positive()) {
		std::cerr << "ERROR : THIS IS A BUG" << std::endl;
	      }
	      continue; // don't examine point at all.
	    }
	    
	    if (CommandlineOptions::verbose()) {
	      std::cerr << "testing(test,testmax)" << test << "," << testMax << std::endl;
	    }
	    
	    if (parent(test,testMax)==vertex) { // check parent
	      vertex = test;
	      vertexMax = testMax;
	      return result;
	    } // check parent
	  } // examine only away from objective
	} // test
      } // loop on leaving edges

    coord = newcoord; // save the current coordinates
  } // while

  vertex = next;
  vertexMax = nextMax;
  return result;
}

// Sets up an incremental Minkowski sum computation
void VPolytopeList::incMinkSumSetup() {

  for (int j = 0; j<numPol(); ++j) {
    (*this)[j].computeAdjacency();
    if (!CommandlineOptions::mute()) {
      std::cerr << "Computed adjacency of " << (j+1) << "th polytope" << std::endl;
    }
  }

  IntegerVector best = IntegerVector(numPol(),0);
  Vector objTemp = Vector((*this)[0].rank(),1);

  Field best_val;


  // All this is necessary to ensure a unique maximum
  // in every polytope, there may be no common maximizer otherwise
  bool objOk = false;
  while (!objOk) {
    objOk = true;
    for (int j=0; objOk && j<numPol(); ++j) {
      best_val = inner_product((*this)[j][0],objTemp);
      best[j] = 0;
      for (size_type i=1; objOk && i<(*this)[j].coldim(); ++i) {
	if (best_val == inner_product((*this)[j][i],objTemp)) {
	  Vector delta = (*this)[j][i];
	  delta.scale(-1);
	  delta.add((*this)[j][best[j]]);
	  if (delta.is_zero()) {
	    std::cerr << "Error : polytope has twice the same vertex" << std::endl;
	    exit(1);
	  }
	  delta.scale(EPSILON);
	  objTemp.add(delta);
	  objOk = false;
	  if (CommandlineOptions::debug()) {
	    std::cerr << "starting again with new objective" << std::endl;
	  }
	} else if (best_val < inner_product((*this)[j][i],objTemp)) {
	  best_val = inner_product((*this)[j][i],objTemp);
	  best[j] = i;
	}
      }
    }
  }

  obj = maximizer(best);
  vertex = best;
  vertexMax = obj;

  if (!CommandlineOptions::mute()) {
    std::cerr << "Setup done" << std::endl;
  }
}


// displays the coordinates of a vector given by its decomposition
void VPolytopeList::outputSetup() {
  std::ofstream adjFile, startFile;
  adjFile.open("adjFile.dat");
  for (size_type i=0; i<numPol(); ++i) {
    (*this)[i].writeAdjacency(adjFile);
  }
  adjFile << obj << std::endl;
  adjFile.close();
  startFile.open("start0.dat");
  startFile << vertex << std::endl;
  startFile << vertexMax << std::endl;
  startFile << 0 << std::endl;
  startFile << 2 << std::endl;
  startFile.close();
}

// returns the coordinates of a vector given by its decomposition
Vector VPolytopeList::output(const IntegerVector sol) const {
  Vector out = Vector((*this)[0][sol[0]]);
  for (int i=1; i<numPol(); ++i) {
    out.add((*this)[i][sol[i]]);
  }
  return out;
}

// displays the coordinates of a vector given by its decomposition
void VPolytopeList::realOutput(const IntegerVector sol) const {
  Vector out = Vector((*this)[0][sol[0]]);
  for (int i=1; i<numPol(); ++i) {
    out.add((*this)[i][sol[i]]);
  }
  std::cout << out;
}

// displays the coordinates of a vector given by its decomposition
void VPolytopeList::realOutput(const IntegerVector sol, std::ostream& os) const {
  Vector out = Vector((*this)[0][sol[0]]);
  for (int i=1; i<numPol(); ++i) {
    out.add((*this)[i][sol[i]]);
  }
  os << out;
}

// displays a vector, without the leading term
void VPolytopeList::print0(const Vector sol) const {
  Vector out = Vector(sol[1]);
  for (int i=2; i<sol.maxindex(); ++i) {
    out.append(sol[i]);
  }
  std::cout << out;
}

// displays a vector, without the leading term
void VPolytopeList::print0(const Vector sol, std::ostream& os) const {
  Vector out = Vector(sol[1]);
  for (int i=2; i<sol.maxindex(); ++i) {
    out.append(sol[i]);
  }
  os << out;
}

// displays a vector
void VPolytopeList::print(const IntegerVector sol) const {
  IntegerVector out = IntegerVector(sol[0]);
  for (int i=1; i<sol.maxindex(); ++i) {
    out.append(sol[i]);
  }
  std::cout << out;
}

// displays a vector
void VPolytopeList::print(const IntegerVector sol, std::ostream& os) const {
  IntegerVector out = IntegerVector(sol[0]);
  for (int i=1; i<sol.maxindex(); ++i) {
    out.append(sol[i]);
  }
  os << out;
}

VPolytope VPolytopeList::removeRedundancy() {
  VPolytope result = VPolytope();

  for (int i=0; i< numPol(); ++i) {
    result.append((*this)[i]);
  }
  result.removeRedundancy();
  return result;
}

// eof VPolytopeList.cc
