/*  This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <assert.h>
#include <fstream>
#include <errno.h>
#include <fcntl.h>
#include "CommandlineOptions.hh"
#include "Array.hh"
#include "VPolytope.hh"
#include "VPolytopeList.hh"
#include "IntegerMatrix.hh"
#include "LPinter.hh"
#include "RSinter.hh"

VPolytopeList polylist;
bool first = true;
size_type proc;
size_type proc_num;
int writeNext;
int readPrev;
int writePrev;
int readNext;
int pipeloop[2];
int pipeloopback[2];
char outFile[15];
std::ofstream out;

void explore(const IntegerVector current, Vector currentMax) {

  VPolytope facets = VPolytope();

  for (int j=0; j<polylist.numPol(); ++j)
    for (int i=0;
	 i<polylist[j].ngbrCount(current[j]);
	 ++i) { // loop on leaving edges

      bool adj;
      Vector edge;
      if (CommandlineOptions::longoutput()) {
	// in this case we must compute adj anyway
	adj = polylist.adj(current,j,i);
	if (adj) { // and add the -edge to the list if adj
	  edge = polylist[j].fej0(current[j],i);
	  edge.scale(-1);
	  facets.append(edge);
	}
      }

      if (CommandlineOptions::longoutput() ?
	  adj && inner_product(edge,polylist.getObj())>0 :
	  (inner_product(polylist[j].fej0(current[j],i),polylist.getObj())<0 &&
	   polylist.adj(current,j,i))) { // examine only away from objective
	IntegerVector test = polylist.neighbour(current,j,i);

	Vector testMax = polylist.maximizer(test);
	if (testMax.maxindex()==0) { // No satisfactory maximizer found
	  // actually, this sould only happen if we choose
	  // to forbid maximizers with negative values
	  if (!CommandlineOptions::positive()) {
	    std::cerr << "ERROR : THIS IS A BUG 2" << std::endl;
	  }
	  continue; // don't examine point at all.
	}

	Matrix pbV = Matrix(testMax.maxindex(),0);

	Vector diff = Vector(testMax);
	diff.scale(-1);
	diff.add(polylist.getObj());

	for (int l=0; l<polylist.numPol(); ++l)
	  for (int k=0;
	       k<polylist[l].ngbrCount(test[l]);
	       ++k)
	    if (inner_product(polylist[l].fej0(test[l],k),diff)>0) {
	      Vector toAdd = Vector(polylist[l].fej0(test[l],k));
	      toAdd.scale(-1);
	      pbV.append(toAdd);
	    }

	RSinter RSV = RSinter(pbV,testMax,diff);
	int limit;

	limit = RSV.shoot();

	if (limit < 1) {
	  fprintf(stderr,"ERROR : THIS IS A BUG\n");
	}

	int firstPol = 0;
	int firstNgbr = 0;
	bool cont = true;

	// find out which is the parent for current vertex
	for (int l=0; cont && l<polylist.numPol(); ++l)
	  for (int k=0; cont && k<polylist[l].ngbrCount(test[l]); ++k)
	    if (inner_product(polylist[l].fej0(test[l],k),diff)>0) {
	      if (--limit == 0) {
		cont = false;
		firstPol = l;
		firstNgbr = k;
	      }
	    }

	if (polylist.neighbour(test,firstPol,firstNgbr)==current) { // check parent
	  int res, message;
	  res = read(readNext,&message,sizeof(message));
	  if (res == -1) {
	    assert(errno == EAGAIN);
	    explore(test,testMax);
	  } else {
	    assert(res == sizeof(message));
	    assert(message == 1);
	    int sendWork = -1;
	    write(writeNext,&sendWork,sizeof(sendWork));
	    for (int l=0;l<polylist.numPol();++l) {
	      write(writeNext,&test[l],sizeof(test[l]));
	    }
	  }
	} // check parent
      } // examine only away from objective
    } // loop on leaving edges

  if (!CommandlineOptions::shortoutput()) {
    polylist.print(current,out);
    out << " : ";
  } else {
    if (!first)
      out << "," << std::endl;
    else
      first = false;
  }
  polylist.realOutput(current,out);
  if (!CommandlineOptions::shortoutput()) {
    out << " : ";
    polylist.print0(currentMax,out);
    if (CommandlineOptions::longoutput()) {
      out << " : ";
      out << facets;
    }
    out << std::endl;
  }
}

int main(const int argc, const char** argv) {
  CommandlineOptions::init(argc, argv);
  proc_num = CommandlineOptions::proc();

  dd_set_global_constants();  // First, this must be called to use cdd

  // Read the polytopes

  if (!polylist.read(std::cin)) {
    fprintf(stderr,"error while reading polytopes\n");
    return 1;
  }

  if (!CommandlineOptions::mute()) {
    fprintf(stderr,"Sum of %d polytopes\n",polylist.numPol());
  }

  polylist.incMinkSumSetup();

  assert(pipe(pipeloop) >= 0);
  assert(pipe(pipeloopback) >= 0);

  proc = 0;
  int pid = 0;
  while ((!pid) && (proc < proc_num-1)) {
    int pipeforknext[2];
    int pipeforkprev[2];
    assert(pipe(pipeforknext) >= 0);
    assert(pipe(pipeforkprev) >= 0);

    pid = fork();
    assert(pid != -1);

    if (pid == 0) {
      readPrev = pipeforknext[0];
      close(pipeforknext[1]);
      writePrev = pipeforkprev[1];
      close(pipeforkprev[0]);
      ++proc;
    } else {
      writeNext = pipeforknext[1];
      close(pipeforknext[0]);
      readNext = pipeforkprev[0];
      close(pipeforkprev[1]);
    }
  }
  if (proc == 0) {
    readPrev = pipeloop[0];
    writePrev = pipeloopback[1];
  } else {
    close(pipeloop[0]);
    close(pipeloopback[1]);
  }
  if (proc == proc_num-1) {
    writeNext = pipeloop[1];
    readNext = pipeloopback[0];
  } else {
    close(pipeloop[1]);
    close(pipeloopback[0]);
  }
  fcntl(readNext,F_SETFL,O_NONBLOCK);

  sprintf(outFile,"result.%d.out",proc);
  out.open(outFile);

  if (CommandlineOptions::shortoutput()) {
    out << "[";
  }

  if (CommandlineOptions::verbose()) {
    fprintf(stderr,"proc %d ready to work\n",proc);
  }

  IntegerVector root = polylist.getBest();
  Vector rootMax = polylist.getObj();
  int res, message;

  int workdone = 0;

  if (proc == 0) {

    explore(root,rootMax);
    ++workdone;
    message = workdone; // signal that proc 0 has nothing to do
    write(writeNext,&message,sizeof(message));

    if (CommandlineOptions::verbose()) {
      fprintf(stderr,"proc %d finished job %d\n",proc,workdone);
    }
  }

  message = 1; // ask for work
  write(writePrev,&message,sizeof(message));

  if (CommandlineOptions::verbose()) {
    fprintf(stderr,"proc %d waiting\n",proc);
  }

  res = read(readPrev,&message,sizeof(message));
  assert(res != -1);
  assert(res == sizeof(message));

  if (CommandlineOptions::verbose()) {
    fprintf(stderr,"proc %d read %d\n",proc,message);
  }

  while (message) {
    if (message == -1) { // signal that previous proc gave work
      IntegerVector work = IntegerVector();
      for (int l=0;l<polylist.numPol();++l) {
	parameter_type temp;
	res = read(readPrev,&temp,sizeof(temp));
	assert(res != -1);
	assert(res == sizeof(temp));
	work.append(temp);
      }
      Vector workMax = polylist.maximizer(work);
      explore(work,workMax);
      ++workdone;

      if (CommandlineOptions::verbose()) {
	fprintf(stderr,"proc %d finished job %d\n",proc,workdone);
      }
      message = 1; // ask for more work
      write(writePrev,&message,sizeof(message));

      if (proc == 0) {
	message = workdone; // signal that proc 0 has nothing to do
	write(writeNext,&message,sizeof(message));
      }
    } else if (message > 0) { // signal sent originally by proc 0
      if (proc == 0) {
	if (message == workdone) { // no work found since then
	  message = 0; // send kill message
	  write(writeNext,&message,sizeof(message));
	} // else ignore, found more work since
      } else {
	write(writeNext,&message,sizeof(message)); // transmit
      }
    } else { // no other messages
      assert(0 == 1); // received an alien message
    }

    if (CommandlineOptions::verbose()) {
      fprintf(stderr,"proc %d waiting\n",proc);
    }

    res = read(readPrev,&message,sizeof(message));
    assert(res != -1);

    assert(res == sizeof(message));

    if (CommandlineOptions::verbose()) {
      fprintf(stderr,"proc %d read %d\n",proc,message);
    }

    if (proc > 0 && message == 0) { // kill message
      write(writeNext,&message,sizeof(message)); // transmit
    }
  } // while(message)

  close(readPrev);
  close(writeNext);
  close(readNext);
  close(writePrev);

  if (CommandlineOptions::shortoutput()) {
    out << "]" << std::endl;
  }
  out.close();

  if (CommandlineOptions::verbose()) {
    fprintf(stderr,"proc %d out after %d jobs done\n",proc,workdone);
  }

  dd_free_global_constants();  // Nice
  return 0;
}
