/*  PyMINKSUM: A Python wrapper to Christophe Weibels MINKSUM C++ library. 
    Copyright (C) 2010  Benjamin Kern <benjamin.kern@ovgu.de>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "_convexsets.h"

Matrix PyArrayToMatrix(PyArrayObject* x) {
  size_t number_of_vertices=PyArray_DIM(x,0);
  size_t dimension_of_vertices=PyArray_DIM(x,1);
  Matrix M(dimension_of_vertices, number_of_vertices);
  for(size_t j=0; j<number_of_vertices; j++){
    for(size_t i=0; i<dimension_of_vertices; i++){
      M[j][i] = *(double *)PyArray_GETPTR2(x,j,i);
    }
  }
  return M;
}

Py::Object MatrixToPyArray(const Matrix& m) {
  PyArrayObject* x = NULL;
//   npy_intp M_Dims[2] = {m.rowdim(), m.coldim()};
  npy_intp M_Dims[2] = {m.coldim(), m.rowdim()};
  x = (PyArrayObject*)PyArray_SimpleNew(2,M_Dims,PyArray_DOUBLE);
  if (x == NULL) {
      Py_XDECREF(x);
      throw Py::ValueError("Problem with creating the return object");
  }
  double* x_ptr = (double*)PyArray_DATA(x);
  for(int j=0; j<M_Dims[0]; j++){
    for(int i=0; i<M_Dims[1]; i++){
      *x_ptr++ = m[j][i];
    }
  }
  return Py::asObject(PyArray_Return(x));
}

void Polytope::init_type() {
  behaviors().name("Polytope Object");
  behaviors().doc("Put Docu in here");
  add_noargs_method("generators", &Polytope::generators, "generators()");
  add_noargs_method("convexhull", &Polytope::convexhull, "convexhull()");
}

Polytope::Polytope(const Matrix& M): vpoly(M), dim_vertices(M.rowdim()) {
//   std::cout << "New Polytope is born" << std::endl;
//   std::cout << vpoly << std::endl;
//   std::cout << dim_vertices << std::endl;
}

Polytope::Polytope(const VPolytope& V): vpoly(V), dim_vertices(V.rowdim()) {
//   std::cout << "New Polytope is born" << std::endl;
//   std::cout << vpoly << std::endl;
//   std::cout << dim_vertices << std::endl;
}

Polytope::~Polytope(){
//   std::cout << "A Polytope died now" << std::endl;
}

Py::Object Polytope::convexhull() {
  VPolytope temp(vpoly);
  dd_set_global_constants();
//   TODO Try catch here possible errors!
  temp.removeRedundancy();
  dd_free_global_constants();
  return Py::asObject(new Polytope(temp));
}

Py::Object Polytope::generators(){
  return MatrixToPyArray(vpoly);
}

Convexsets::Convexsets() : Py::ExtensionModule<Convexsets>("_convexsets"){
  Polytope::init_type();
  add_varargs_method("Polytope", &Convexsets::new_polytope, "Polytope Object");
  add_varargs_method("Minksum", &Convexsets::minksum, "Minksum((P1,..,Pn))");
  // TODO Adapt docu here.
  initialize("Input documentation of Convexsets module in here ");
}

Py::Object Convexsets::minksum(const Py::Tuple &args){
  args.verify_length(1);
  // TODO Check that at least 2 Elements are in the list.
  Py::Tuple polytuple=args[0];
  if (!Polytope::check(polytuple[0].ptr()))
    throw Py::TypeError("Provide a tuple of polytope objects");
  if(polytuple.size()<2)
    throw Py::TypeError("At least 2 polytope objects are needed");
  Polytope* p = static_cast<Polytope*>(polytuple[0].ptr());
  size_t dim_vertices = p->get_dim_vertices();
  VPolytopeList polylist(p->get_vpolytope());
  for (size_t i=1; i<polytuple.size(); i++) {
    if (!Polytope::check(polytuple[i].ptr()))
      throw Py::TypeError("Provide a tuple of polytope objects");
    Polytope* p = static_cast<Polytope*>(polytuple[i].ptr());
    if (dim_vertices != p->get_dim_vertices())
      throw Py::TypeError("Dimension mismatch between the polytopes");
    polylist.append(p->get_vpolytope());
  }
  dd_set_global_constants();
//   TODO Try catch here possible errors!
  polylist.incMinkSumSetup();
  Matrix temp(polylist.incExploreStepFast().coord());
  while (polylist.hasMoreVertices()) {
      temp.append(polylist.incExploreStepFast().coord());
  }
  dd_free_global_constants();
  return Py::asObject(new Polytope(temp));
}

Py::Object Convexsets::new_polytope(const Py::Tuple &args){
 // TODO Check if one or 3-arguments are given and then create either a H or an
 // V Polytope.
  args.verify_length(1);
  PyArrayObject* x = NULL;
  x = (PyArrayObject*)PyArray_ContiguousFromObject(args[0].ptr(), 
                                                   PyArray_DOUBLE, 2, 2);
  if (x == NULL) {
      Py_XDECREF(x);
      throw Py::ValueError("x needs to be a 2D array");
  }

  VPolytope vpoly=PyArrayToMatrix(x);
//   Matrix data=PyArrayToMatrix(x);

  Py_XDECREF(x);
  return Py::asObject(new Polytope(vpoly));
}

extern "C" void init_convexsets(){
  static Convexsets* Convexsets_module = new Convexsets;
  import_array();
}
